unit qiFormat;
{
Copyright 1996 by Justin Turberville
EMail: justint@cyberjoe.co.za

TQIFormatter is designed to propduce both SQL-compliant and user-orintated
versions of the query action data stored in A TQueryInfo component.
Sub-type this component to customise how the data is interpreted (for
various SQL and QBE standards, and database types and drivers, or just to
introduce improved  routines) and for international use (other languages
for user version routines).
See DemoUnit.pas to see how to ensure that TQIFormatter includes required
bridging joins in multi-table queries (application-specific logic).
}
interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls, DB,
  qiField, qiQuery, qiUpdate;

type
  TSQLClauseEvent = procedure(Sender: TObject; var Clause: string) of object;

  TUserText = class(TPersistent)
  private
    FSelect, FFrom, FWhere, FGroup, FOrder, FModify, FInsert, FDelete,
      FNoWhere, FNoGroup, FNoOrder, FDefaultSelect: string;
  public
    constructor Create;
  published
    property Select: string read FSelect write FSelect;
    property From: string read FFrom write FFrom;
    property Where: string read FWhere write FWhere;
    property Group: string read FGroup write FGroup;
    property Order: string read FOrder write FOrder;
    property Modify: string read FModify write FModify;
    property Insert: string read FInsert write FInsert;
    property Delete: string read FDelete write FDelete;
    property NoWhere: string read FNoWhere write FNoWhere;
    property NoGroup: string read FNoGroup write FNoGroup;
    property NoOrder: string read FNoOrder write FNoOrder;
    property DefaultSelect: string read FDefaultSelect write FDefaultSelect;
  end;

  TQIFormatter = class(TComponent)
  private
    FAbout: TQIAboutProperty;
    FIndent: Integer;
    FPrefix: Boolean;
    FUserText: TUserText;
    FShowFrom: Boolean;
    FShowGroup: Boolean;
    FQueryInfo: TQIQueryInfo;
    FUpdateInfo: TQIUpdateInfo;
    FSQLStandard: TSQLStandard;
    FAddTableName: Boolean;
    FDefaultSelect: Boolean;
    FOnSelectClause, FOnFromClause, FOnJoinClause, FOnFilterClause,
      FOnGroupClause, FOnOrderClause, FOnHavingClause: TSQLClauseEvent;
    procedure SetQueryInfo(Value: TQIQueryInfo);
    procedure SetUpdateInfo(Value: TQIUpdateInfo);
    procedure SetIndent(Value: Integer);
    function IndentStr: string;
    function WithNester(Filter: string; Nester: TNestingType): string;
    function UserStr(AddText, BodyText: string): string;
    function WithTableName(Line: string; Field: TFieldItem; ToInsert: Boolean): string;
  protected
    procedure CheckQueryInfo;
    procedure CheckUpdateInfo;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    function SQLSelectClause: string;
    function SQLFromClause: string;
    function SQLJoinClause: string;
    function SQLFilterClause(Kind: TFilterKind): string;
    function SQLHavingClause: string;
    function SQLGroupClause: string;
    function SQLOrderClause: string;
    function UserSelectClause: string;
    function UserFromClause: string;
    function UserFilterClause(Kind: TFilterKind): string;
    function UserGroupClause: string;
    function UserOrderClause: string;
    procedure SQLFilterStrings(Kind: TFilterKind; Strings: TStrings);
    procedure UserFilterStrings(Kind: TFilterKind; Strings: TStrings);
    procedure SQLQueryScript(Strings: TStrings);
    procedure UserQueryScript(Strings: TStrings);
    procedure SQLUpdateScript(Strings: TStrings);
    procedure UserUpdateScript(Strings: TStrings);
  published
    property About: TQIAboutProperty read FAbout write FAbout stored False;
    property DefaultSelect: Boolean read FDefaultSelect write FDefaultSelect default False;
    property Prefix: Boolean read FPrefix write FPrefix default True;
    property QueryInfo: TQIQueryInfo read FQueryInfo write SetQueryInfo;
    property UpdateInfo: TQIUpdateInfo read FUpdateInfo write SetUpdateInfo;
    property SQLStandard: TSQLStandard read FSQLStandard write FSQLStandard default sqDefault;
    property AddTableName: Boolean read FAddTableName write FAddTableName default False;
    property Indent: Integer read FIndent write SetIndent default 3;
    property UserText: TUserText read FUserText write FUserText;
    property ShowFrom: Boolean read FShowFrom write FShowFrom default False;
    property ShowGroup: Boolean read FShowGroup write FShowGroup default True;
    property OnSelectClause: TSQLClauseEvent read FOnSelectClause write FOnSelectClause;
    property OnFromClause: TSQLClauseEvent read FOnFromClause write FOnFromClause;
    property OnJoinClause: TSQLClauseEvent read FOnJoinClause write FOnJoinClause;
    property OnFilterClause: TSQLClauseEvent read FOnFilterClause write FOnFilterClause;
    property OnGroupClause: TSQLClauseEvent read FOnOrderClause write FOnGroupClause;
    property OnOrderClause: TSQLClauseEvent read FOnOrderClause write FOnOrderClause;
    property OnHavingClause: TSQLClauseEvent read FOnHavingClause write FOnHavingClause;
  end;

implementation

constructor TUserText.Create;
begin
  inherited Create;
  FSelect := 'Show';
  FFrom   := '  from';
  FWhere  := 'Filter where';
  FGroup  := 'Group by';
  FOrder  := 'Order by';
  FModify := 'Update';
  FInsert := 'Insert into';
  FDelete := 'Delete from';
  FNoWhere := '';
  FNoGroup := '';
  FNoOrder := 'Default order';
  FDefaultSelect := 'all fields';
end;

constructor TQIFormatter.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FUserText := TUserText.Create;
  FIndent := 3;
  FPrefix := True;
{ FDefaultSelect := True;}
  FSQLStandard := sqDefault;
  FShowGroup := True;
end;

destructor TQIFormatter.Destroy;
begin
  FUserText.Free;
  inherited Destroy;
end;

procedure TQIFormatter.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if Operation = opRemove then
    if AComponent = FQueryInfo then FQueryInfo := nil else
    if AComponent = FUpdateInfo then FUpdateInfo := nil;
end;

procedure TQIFormatter.SetQueryInfo(Value: TQIQueryInfo);
begin
  FQueryInfo := Value;
  if Value <> nil then
    Value.FreeNotification(Self);
end;

procedure TQIFormatter.SetUpdateInfo(Value: TQIUpdateInfo);
begin
  FUpdateInfo := Value;
  if Value <> nil then
    Value.FreeNotification(Self);
end;

procedure TQIFormatter.SetIndent(Value: Integer);
begin
  if Value < 0 then FIndent := 0
  else if Value > 8 then FIndent := 8
  else FIndent := Value;
end;

procedure TQIFormatter.CheckQueryInfo;
begin
  if FQueryInfo = nil then
    raise EComponentError.Create(sNoQueryInfo);
end;

procedure TQIFormatter.CheckUpdateInfo;
begin
  if FUpdateInfo = nil then
    raise EComponentError.Create(sNoUpdateInfo);
end;

function TQIFormatter.IndentStr: string;
begin
  Result := Format('%' + IntToStr(FIndent) + 's', ['']);
end;

function TQIFormatter.WithNester(Filter: string; Nester: TNestingType): string;
begin
  Result := Filter;
  if (Result <> '') and (Nester > ntNone) then
    if Nester = ntOpening then
      if Result[1] in ['A', 'a'] then
        Insert(' (', Result, 4) else      {insert after AND}
        Insert(' (', Result, 3)           {insert after OR}
    else Result := Result + ' )'          {add to end}
end;

function TQIFormatter.UserStr(AddText, BodyText: string): string;
begin
  if BodyText = '' then Result := AddText
  else if FPrefix then Result := AddText + ' ' + BodyText
  else Result := BodyText;
end;

function TQIFormatter.WithTableName(Line: string; Field: TFieldItem;
  ToInsert: Boolean): string;
var
  i: Integer;
begin
  Result := Line;
  if ToInsert and (Field <> nil) then begin
    i := Pos(Field.UserName, Result);
    if i > 0 then
      Insert(Field.Table.UserName + ' ', Result, i);
  end;
end;

function TQIFormatter.SQLSelectClause: string;
var
  i: Integer;
begin
  Result := '';
  CheckQueryInfo;
  with FQueryInfo do
    if Selections.Count = 0 then  {if Selections not used}
      if FDefaultSelect then
        for i := 0 to Fields.Count-1 do begin     {then include by default -}
          with Fields[i] do
            if (fuSelect in Usage) or FieldRequired(Fields[i]) then
              Result := Result + SQLNameStr(True) + ', ';
        end
      else Exit
    else begin                         {else do it the hard way -}
      for i := 0 to Selections.Count-1 do
        Result := Result + Selections[i].SQLSelectStr + ', ';
      for i := 0 to Fields.Count-1 do   {get any required fields not in -}
        if FieldRequired(Fields[i])  and
          (Selections.ByFieldAndAggregate(Fields[i], qaNone) = nil) then
          Result := Result + Fields[i].SQLNameStr(True) + ', ';
    end;
  System.Delete(Result, Length(Result) -1, 2);         {remove last comma}
  if QueryInfo.Selections.Distinct then
    Result := 'DISTINCT ' + Result;
  if FPrefix then Result := 'SELECT ' + Result;
  if Assigned(FOnSelectClause) then FOnSelectClause(Self, Result);
end;

function TQIFormatter.UserSelectClause: string;
var
  i: Integer;
begin
  Result := '';
  CheckQueryInfo;
  with FQueryInfo do begin
    if Selections.Count = 0 then      {if Selections not used -}
      if FDefaultSelect then
        Result := FUserText.DefaultSelect  {presumed to be all defined fields}
      else Exit
    else begin                   {else only add specifically selected fields -}
      for i := 0 to Selections.Count-1 do
        Result := Result + WithTableName(Selections[i].UserSelectStr,
          Selections[i].Field, FAddTableName) + ', ';
      System.Delete(Result, Length(Result) -1, 2);
    end;
    if Selections.Distinct then
      Result := 'distinct ' + Result;
  end;
  Result := UserStr(FUserText.Select, Result);
end;

function TQIFormatter.SQLFromClause: string;
var
  i: Integer;
begin
  Result := '';
  CheckQueryInfo;
  with FQueryInfo do
    for i := 0 to Tables.Count-1 do
      if TableUsed(Tables[i]) then Result :=
        Result + StripExt(Tables[i].SQLName)+ ' ' + Tables[i].Alias + ', ';
  if Result <> '' then begin
    System.Delete(Result, Length(Result)-1, 2);
    if FPrefix then Result := 'FROM ' + Result;
  end;
  if Assigned(FOnFromClause) then FOnFromClause(Self, Result);
end;

function TQIFormatter.UserFromClause: string;
var
  i: Integer;
begin
  Result := '';
  CheckQueryInfo;
  with FQueryInfo do
    for i := 0 to Tables.Count-1 do
      if TableUsed(Tables[i]) then
        Result := Result + Tables[i].UserName + ', ';
  if Result <> '' then begin
    System.Delete(Result, Length(Result)-1, 2);
    Result := UserStr(FUserText.From, Result);
  end;
end;

function TQIFormatter.SQLJoinClause: string;
var
  i: Integer;
begin
  Result := '';
  CheckQueryInfo;
  with FQueryInfo do        {include all joins defined between used tables -}
    for i := 0 to Joins.Count-1 do
      with Joins[i] do
        if TableUsed(Table1) and TableUsed(Table2) then
           Result := Result + SQLJoinStr + ' AND ';
  if Result <> '' then begin
    System.Delete(Result, Length(Result)-4, 5);        {delete last ' AND '}
    if FPrefix then Result := 'WHERE ' + Result;
  end;
  if Assigned(FOnJoinClause) then FOnJoinClause(Self, Result);
end;

function TQIFormatter.SQLFilterClause(Kind: TFilterKind): string;
var
  i: Integer;
  filters: TFilterList;
begin
  Result := '';
  if Kind = fkQuery then begin
    CheckQueryInfo;
    filters := FQueryInfo.Filters;
  end
  else begin
    CheckUpdateInfo;
    filters := FUpdateInfo.Filters;
  end;
  for i := 0 to filters.Count-1 do
    if filters[i].Aggregate = qaNone then
      Result := Result + WithNester(filters[i].SQLFilterStr,
        QueryInfo.Filters.NestingType(filters[i])) + ' ';
  if Result <> '' then begin
    System.Delete(Result, Length(Result), 1);    {delete last space}
    if Result[1] = 'A' then
      System.Delete(Result, 1, 4) else         {delete first AND}
      System.Delete(Result, 1, 3);             {delete first OR}
    Result := '(' + Result + ')';    {precaution for un-nested OR's after join}
    if FPrefix then Result := 'WHERE ' + Result;
  end;
  if Assigned(FOnFilterClause) then FOnFilterClause(Self, Result);
end;

function TQIFormatter.UserFilterClause(Kind: TFilterKind): string;
var
  i: Integer;
  s: string;
  filters: TFilterList;
begin
  Result := '';
  if Kind = fkQuery then begin
    CheckQueryInfo;
    filters := FQueryInfo.Filters;
  end
  else begin
    CheckUpdateInfo;
    filters := FUpdateInfo.Filters;
  end;
  for i := 0 to filters.Count-1 do begin
    s := WithNester(filters[i].UserFilterStr,
      QueryInfo.Filters.NestingType(filters[i]));
    s := WithTableName(s, filters[i].Field, FAddTableName);
    if filters[i].OpField <> nil then
      s := WithTableName(s, filters[i].OpField, FAddTableName);
    Result := Result + s + ' ';
  end;
  if Result <> '' then begin
    System.Delete(Result, Length(Result), 1);    {delete last space}
    if Result[1] = 'a' then
      System.Delete(Result, 1, 4) else         {'and '}
      System.Delete(Result, 1, 3);             {'or '}
    System.Delete(Result, Length(Result), 1);  {' '}
    Result := UserStr(FUserText.Where, Result);
  end
  else Result := UserStr(FUserText.NoWhere, '');
end;

function TQIFormatter.SQLHavingClause: string;
var
  i: Integer;
begin
  Result := '';
  CheckQueryInfo;
  with QueryInfo do
    for i := 0 to Filters.Count-1 do
      if Filters[i].Aggregate > qaNone then
        Result := Result + WithNester(filters[i].SQLFilterStr,
          QueryInfo.Filters.NestingType(filters[i])) + ' ';
  if Result <> '' then begin
    System.Delete(Result, Length(Result), 1);    {delete last space}
    if Result[1] = 'A' then
      System.Delete(Result, 1, 4) else         {delete first AND}
      System.Delete(Result, 1, 3);             {delete first OR}
    if FPrefix then Result := 'HAVING ' + Result;
  end;
  if Assigned(FOnHavingClause) then FOnHavingClause(Self, Result);
end;

function TQIFormatter.SQLGroupClause: string;
var
  i: Integer;
begin
  Result := '';
  CheckQueryInfo;
  with FQueryInfo do begin
    for i := 0 to Grouping.Count-1 do
      Result := Result + Grouping[i].SQLGroupStr + ', ';
    if Grouping.Count > 0 then
      for i := 0 to Fields.Count-1 do   {get any required fields not in -}
        if FieldRequired(Fields[i]) and (Grouping.ByField(Fields[i]) = nil) then
          Result := Result + Fields[i].SQLNameStr(False) + ', ';
  end;
  if Result <> '' then begin
    System.Delete(Result, Length(Result) -1, 2);
    if FPrefix then Result := 'GROUP BY ' + Result;
  end;
  if Assigned(FOnGroupClause) then FOnGroupClause(Self, Result);
end;

function TQIFormatter.UserGroupClause: string;
var
  i: Integer;
begin
  Result := '';
  CheckQueryInfo;
  with FQueryInfo do
    for i := 0 to Grouping.Count-1 do
      Result := Result + WithTableName(Grouping[i].UserGroupStr,
        Grouping[i].Field, FAddTableName) + ', ';
  if Result <> '' then begin
    System.Delete(Result, Length(Result) -1, 2);
    Result := UserStr(FUserText.Group, Result);
  end
  else Result := UserStr(FUserText.NoGroup, '');
end;

function TQIFormatter.SQLOrderClause: string;
  function SelectionIndex(F: TFieldItem; A: TQueryAggregate): Integer;
  var
    i: Integer;
  begin
    Result := -1;
    for i := 0 to QueryInfo.Selections.Count-1 do
      with QueryInfo.Selections[i] do
        if (Field = F) and (Aggregate = A) then begin
          Result := i;
          Break;
        end;
  end;
var
  i, n: Integer;
begin
  Result := '';
  CheckQueryInfo;
  with FQueryInfo do
    for i := 0 to Ordering.Count-1 do
      with Ordering[i] do begin
        if Aggregate = qaNone then
          n := -1 else
          n := SelectionIndex(Field, Aggregate);
        Result := Result + SQLOrderStr(n) + ', ';
      end;
  if Result <> '' then begin
    System.Delete(Result, Length(Result) -1, 2);
    if FPrefix then Result := 'ORDER BY ' + Result;
  end;
  if Assigned(FOnOrderClause) then FOnOrderClause(Self, Result);
end;

function TQIFormatter.UserOrderClause: string;
var
  i: Integer;
begin
  Result := '';
  CheckQueryInfo;
  with FQueryInfo do
    for i := 0 to Ordering.Count-1 do
      Result := Result + WithTableName(Ordering[i].UserOrderStr,
        Ordering[i].Field, FAddTableName) + ', ';
  if Result <> '' then begin
    System.Delete(Result, Length(Result) -1, 2);
    Result := UserStr(FUserText.Order, Result);
  end
  else Result := UserStr(FUserText.NoOrder, '');
end;

procedure TQIFormatter.SQLFilterStrings(Kind: TFilterKind; Strings: TStrings);
var
  i, n: Integer;
  s: string;
  filters: TFilterList;
begin
  Strings.Clear;
  if Kind = fkQuery then begin
    CheckQueryInfo;
    filters := FQueryInfo.Filters;
  end
  else begin
    CheckUpdateInfo;
    filters := FUpdateInfo.Filters;
  end;
  for i := 0 to filters.Count-1 do
    with filters[i] do
      if Aggregate = qaNone then begin
        s := WithNester(SQLFilterStr, QueryInfo.Filters.NestingType(filters[i]));
        if Kind = fkUpdate then begin        {remove table alias qualifier -}
          n := Pos(Field.Table.Alias, s);
          if n > 0 then
            System.Delete(s, n, Length(Field.Table.Alias)+1);
        end;
        Strings.Add(s);
      end;
  if Strings.Count > 0 then begin
    s := Strings[0];
    if s[1] = 'A' then
      System.Delete(s, 1, 4) else         {delete first AND}
      System.Delete(s, 1, 3);             {delete first OR}
    Strings[0] := '( ' + s;   {precaution for un-nested OR's after join -}
    Strings[Strings.Count-1] := Strings[Strings.Count-1] + ' )';
    if FPrefix then
      Strings[0] := 'WHERE ' + Strings[0];
    if FIndent > 0 then
      for i := 1 to Strings.Count-1 do           {add indent -}
        Strings[i] := IndentStr + Strings[i];
  end;
end;

procedure TQIFormatter.UserFilterStrings(Kind: TFilterKind; Strings: TStrings);
var
  i: Integer;
  s: string;
  filters: TFilterList;
begin
  Strings.Clear;
  if Kind = fkQuery then begin
    CheckQueryInfo;
    filters := FQueryInfo.Filters;
  end
  else begin
    CheckUpdateInfo;
    filters := FUpdateInfo.Filters;
  end;
  for i := 0 to filters.Count-1 do begin
    s := WithNester(filters[i].UserFilterStr,
      QueryInfo.Filters.NestingType(filters[i]));
    s := WithTableName(s, filters[i].Field, FAddTableName);
    if filters[i].OpField <> nil then
      s := WithTableName(s, filters[i].OpField, FAddTableName);
    Strings.Add(s);
  end;
  if Strings.Count > 0 then begin
    s := Strings[0];
    if s[1] = 'a' then
      System.Delete(s, 1, 4) else         {delete first AND}
      System.Delete(s, 1, 3);             {delete first OR}
    Strings[0] := UserStr(FUserText.Where, s);
    if FIndent > 0 then
      for i := 1 to Strings.Count-1 do           {add indent -}
        Strings[i] := IndentStr + Strings[i];
  end
  else begin
    s := UserStr(FUserText.NoWhere, '');
    if s <> '' then Strings.Add(s);
  end;
end;

procedure TQIFormatter.SQLQueryScript(Strings: TStrings);
var
  s: string;
  i: Integer;
  l: TStringList;
  p, b: Boolean;
begin
  CheckQueryInfo;
  p := FPrefix;
  with Strings, FQueryInfo do
  try
    l := TStringList.Create;
    BeginUpdate;
    Clear;
    FPrefix := True;
    s := SQLSelectClause;
    if s = '' then Exit;
    Add(s);
    Add(SQLFromClause);
    s := SQLJoinClause;
    b := s <> '';
    if b then Add(s);
    SQLFilterStrings(fkQuery, l);
    if b and (l.Count > 0) then begin    {join clause included }
      s := l[0];
      System.Delete(s, 1, 5);        {so make WHERE into AND -}
      l[0] := IndentStr + 'AND' + s;
    end;
    for i := 0 to l.Count-1 do
      Add(l[i]);
    s := SQLGroupClause;
    if s <> '' then Add(s);
    s := SQLHavingClause;
    if s <> '' then Add(s);
    s := SQLOrderClause;
    if s <> '' then Add(s);
  finally
    EndUpdate;
    FPrefix := p;
    l.Free;
  end;
end;

procedure TQIFormatter.UserQueryScript(Strings: TStrings);
var
  s: string;
  i: Integer;
  p: Boolean;
begin
  CheckQueryInfo;
  p := FPrefix;
  try
    Strings.BeginUpdate;
    Strings.Clear;
    FPrefix := True;
    s := UserSelectClause;
    if s = '' then Exit;
    UserFilterStrings(fkQuery, Strings);
    Strings.Insert(0, s);
    s := UserFromClause;
    if FShowFrom and (s <> '') then
      Strings.Insert(1, s);
    s := UserGroupClause;
    if FShowGroup and (s <> '') then
      Strings.Add(s);
    Strings.Add(UserOrderClause);
  finally
    Strings.EndUpDate;
    FPrefix := p;
  end;
end;

procedure TQIFormatter.SQLUpdateScript(Strings: TStrings);
var
  i: Integer;
  s: string;
  p: Boolean;
begin
  CheckUpdateInfo;
  p := FPrefix;
  FPrefix := True;
  with FUpdateInfo do
  try
    Strings.BeginUpdate;
    Strings.Clear;
    case UpdateKind of
      ukModify: begin
        if Updates.Count = 0 then Exit;
        SQLFilterStrings(fkUpdate, Strings);
        s := 'SET ';
        for i := 0 to Updates.Count-1 do
          s := s + Updates[i].SQLModifyStr + ', ';
        System.Delete(s, Length(s)-1, 2);
        Strings.Insert(0, s);
        Strings.Insert(0, 'UPDATE ' + StripExt(UpdateTable.SQLName));
      end;
      ukInsert: begin
        if Updates.Count = 0 then Exit;
        Strings.Add ('INSERT INTO ' + StripExt(UpdateTable.SQLName));
        s := ' (';
        for i := 0 to Updates.Count-1 do               {add field names -}
          s := s + Updates[i].Field.SQLName + ', ';
        System.Delete(s, Length(s)-1, 2);
        Strings.Add(s + ')');
        s := 'VALUES (';
        for i := 0 to Updates.Count-1 do               {add field values -}
          with Updates[i] do
            s := s + SQLValueStr(Field.FieldType, Value) + ', ';
        System.Delete(s, Length(s)-1, 2);
        Strings.Add(s + ')');
      end;
      ukDelete: begin
        if UpdateTable = nil then Exit;
        SQLFilterStrings(fkUpdate, Strings);
        Strings.Insert(0, 'DELETE FROM ' + StripExt(UpdateTable.SQLName));
      end;
    end;
  finally
    Strings.EndUpdate;
    FPrefix := p;
  end;
end;

procedure TQIFormatter.UserUpdateScript(Strings: TStrings);
var
  i: Integer;
  s: string;
  p: Boolean;
begin
  CheckUpdateInfo;
  p := FPrefix;
  FPrefix := True;
  with FUpdateInfo do
  try
    Strings.BeginUpdate;
    Strings.Clear;
    if UpdateTable = nil then Exit;
    case UpdateKind of
      ukModify:
        if Updates.Count > 0 then begin
          UserFilterStrings(fkUpdate, Strings);
          for i := 0 to Updates.Count-1 do
            Strings.Insert(i, IndentStr + WithTableName(
              Updates[i].UserModifyStr, Updates[i].Field, AddTableName));
          Strings.Insert(0, UserStr(FUserText.Modify, UpdateTable.UserName));
        end;
      ukInsert:
        if Updates.Count > 0 then begin
          Strings.Add(UserStr(FUserText.Insert, UpdateTable.UserName));
          for i := 0 to Updates.Count-1 do
            Strings.Add(IndentStr + WithTableName(
              Updates[i].UserModifyStr, Updates[i].Field, AddTableName));
        end;
      ukDelete:
        if UpdateTable <> nil then begin
          UserFilterStrings(fkUpdate, Strings);
          Strings.Insert(0, UserStr(FUserText.Delete, UpdateTable.UserName));
        end;
    end;
  finally
    Strings.EndUpdate;
    FPrefix := p;
  end;
end;

end.
